/* Electronic Handball */

#include <stdio.h>
#include <curses.h>
#include <sys/time.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "server.h"
#include "sock.h"

#ifdef SAAGAR
#define MSG_WAITALL 0
#endif

void drawBox(WINDOW *win);
void drawScore(WINDOW *win);
void drawPaddle(WINDOW *win, int paddlexold, int paddlex, int paddley);
void drawScreen(WINDOW *win, int ballxold, int ballyold, int ballx, int bally);
int hitPaddle(int *ballx, int *ballx_vel, int *bally, int *bally_vel, 
	       int paddlex, int paddley);
void setAlarm(long milleseconds);
void onAlarm(int sig);
void getMove(WINDOW *win, int *movex);
int moveBall(int *ballxold, int *ballyold, int *ballx, int *bally, 
	     int *ballx_vel, int *bally_vel, int paddlex, int paddley);
void movePaddle(int *paddlexold, int *paddlex, int movex);
void Usage(void);
void cleanUp(int sig);
void getState();
void sendPaddleMove(int paddlex);
void sendBallMove(int ballx, int bally, int ballx_vel, int bally_vel);
void sendSwitchTurn();
void sendScore();

#define VERSION         1.0
#define BALL_CHAR       '*'
#define PADDLE_CHAR     '='
#define BEEP_CHAR       ''
#define MIN_X		21
#define MAX_X		59
#define MIN_Y		2
#define MAX_Y 		21
#define DELAY		75
#define PADDLE_SIZE     2

/* global variables */
int id, other_id;
int wall = 0;
int sock = -1;
int my_pid = -1;
struct Table table;

int main(int argc, char *argv[]) {
  WINDOW *main_window;		/* curses window */
  int ballxold, ballyold;
  int movex;
  int paddlexold;
  int doMove = 1;		/* move every other time */
  int created;			/* if 1st one into pong game */
  char name[MAX_NAME] = "";	/* name */
  int i,j;			/* loop index */
  int didScore = 0;
  struct textMessage msg;
  int bytes;
  int port = 0;
  char *server = NULL;
  int c;
  extern int optind, opterr;
  extern char *optarg;

  opterr = 1; /* set to 0 to disable error message */

  /* required: one of {-b or -p}, optional: -i<name> and/or -o<name> */
  while ((c = getopt (argc, argv, "hp:s:")) != EOF) {
    switch (c) {
    case '?': /* returns ? if there is a getopt error */
    case 'h':
      Usage();
      break;
    case 'p':
      port = atoi(optarg);
      break;
    case 's':
      server = optarg;
      break;
    }

  }
  if (!port) {
    printf("You must specify a port.\n");
    Usage();
  }
  if (!server) {
    printf("You must specify a server.\n");
    Usage();
  }

  printf("hanball client activated.\n");

  /* set up interrupt handlers */
  signal(SIGALRM, onAlarm);
  signal(SIGINT, cleanUp);	/* Ctrl-C to exit */
  
  /* get name */
  system("clear");
  printf("** Welcome to Electronic Handball ** \n\n");
  printf("Enter your name: ");
  while (strlen(name) == 0)
    fgets(name, MAX_NAME, stdin);
  name[strlen(name)-1] = '\0';	/* remove \n */

  /* connect to server */
  printf("Connecting to %s...\n", server);
  sock = sockconnect(server, port);
  if (sock == -1) {
    perror("sockconnect");
    cleanUp(0);
  }

  /* send name */
  printf("Sending name to join game...", msg.note);
  msg.mtype = MESSAGE;
  strcpy(msg.note, name);
  bytes = send(sock, &msg, sizeof(struct textMessage), 0);
  if (bytes == -1)
    perror("send");
  bytes = recv(sock, &msg, sizeof(struct textMessage), MSG_WAITALL);
  if (msg.mtype != MESSAGE) {
    printf("Unexpected message type: %d\n", msg.mtype);
    cleanUp(0);
  }
  printf("Received: %s\n", msg.note);
  my_pid = atoi(msg.note);
  if (my_pid < 0) {
    printf("Game must be full.\n");
    cleanUp(0);
  }
  printf("My number: '%d'\n", my_pid);


  /* initialize curses stuff */  
  initscr();
  main_window = newwin(0,0,0,0);
  leaveok(main_window,TRUE);	/* turn off cursor? */
  keypad(stdscr, TRUE);		/* get arrow keys */
  cbreak();			/* don't wait for \n */
  noecho();			/* don't echo keys pressed */
  nodelay(main_window, TRUE);	/* don't block on getch */
  
  /* draw starting stuff */
  drawBox(main_window);
  drawScore(main_window);
  
  /* initialize */
  getState();
  ballxold = table.ballx;
  ballyold = table.bally;
  movex = 0;
  paddlexold = table.paddlex;

  /* loop until exit on Ctrl-C */
  while (1) { 

    /* wait for alarm to go off */
    setAlarm(DELAY);
    pause();		

    /* get current state */
    getState();

    /* do moves */
    if (table.turn == my_pid) {
      if (doMove) {
	didScore = moveBall(&ballxold, &ballyold, &table.ballx, &table.bally, 
			    &table.ballx_vel, &table.bally_vel, 
			    table.paddlex, table.paddley);
	if (!didScore)
	  sendBallMove(table.ballx, table.bally, 
		       table.ballx_vel, table.bally_vel);
	doMove = 0;
      } else
	doMove = 1;
      getMove(main_window, &movex);
      movePaddle(&paddlexold, &table.paddlex, movex);

      /* send paddle move */
      sendPaddleMove(table.paddlex);
    
    }

    /* get current state */
    getState();

    /* draw */
    /* hack to clear the screen */
    for (i=MIN_Y; i<=MAX_Y; i++)
      for(j = MIN_X; j<=MAX_X; j++)
	mvwaddch(main_window, i, j, ' ');
    drawScreen(main_window, ballxold, ballyold, table.ballx, table.bally);
    drawPaddle(main_window, paddlexold, table.paddlex, table.paddley);

    /* reset */
    ballxold = table.ballx;
    ballyold = table.bally;
    paddlexold = table.paddlex;

    drawScore(main_window);
    wrefresh(main_window);
    
  }
}

/* get the pong state from the server */
void getState() {
  struct Message buff;
  struct textMessage msg;
  extern struct Table table;
  int bytes;

  buff.mtype = SENDGAME;
  bytes = send(sock, &buff, sizeof(struct Message), 0);
  if (bytes == -1) {
    perror("send");
    cleanUp(0);
  }
  bytes = recv(sock, &table, sizeof(struct Table), MSG_WAITALL);

  return;
}

/* send the pong state to the server */
void sendPaddleMove(int paddlex) {
  struct Message buff;
  struct textMessage msg;
  int bytes;

  buff.mtype = MOVEPADDLE;
  buff.x = paddlex;
  buff.y = table.paddley;
  bytes = send(sock, &buff, sizeof(struct Message), 0);
  if (bytes == -1) {
    perror("send");
    cleanUp(0);
  }
  bytes = recv(sock, &msg, sizeof(struct textMessage), MSG_WAITALL);
  if (msg.mtype != MESSAGE) {
    printf("Unexpected message type: %d\n", msg.mtype);
    cleanUp(0);
  }
  if (strcmp(msg.note, "Ok") != 0) {
    printf("Server says: '%s'\n", msg.note);
    cleanUp(0);
  }
}


/* send the pong state to the server */
void sendBallMove(int ballx, int bally, int ballx_vel, int bally_vel) {
  struct Message buff;
  struct textMessage msg;
  int bytes;

  buff.mtype = MOVEBALL;
  buff.x = ballx;
  buff.y = bally;
  buff.x_vel = ballx_vel;
  buff.y_vel = bally_vel;
  bytes = send(sock, &buff, sizeof(struct Message), 0);
  if (bytes == -1) {
    perror("send");
    cleanUp(0);
  }
  bytes = recv(sock, &msg, sizeof(struct textMessage), MSG_WAITALL);
  if (msg.mtype != MESSAGE) {
    printf("Unexpected message type: %d\n", msg.mtype);
    cleanUp(0);
  }
  if (strcmp(msg.note, "Ok") != 0) {
    printf("Server says: '%s'\n", msg.note);
    cleanUp(0);
  }
}

/* send a switch turn command */
void sendSwitchTurn() {
  struct textMessage msg;
  struct Message buff;
  int bytes;

  buff.mtype = SWITCHTURN;
  bytes = send(sock, &buff, sizeof(struct Message), 0);
  if (bytes == -1) {
    perror("send");
    cleanUp(0);
  }
  bytes = recv(sock, &msg, sizeof(struct textMessage), MSG_WAITALL);
  if (msg.mtype != MESSAGE) {
    printf("Unexpected message type: %d\n", msg.mtype);
    cleanUp(0);
  }
  if (strcmp(msg.note, "Ok") != 0) {
    printf("Server says: '%s'\n", msg.note);
    cleanUp(0);
  }
}

/* send a score command */
void sendScore(void) {
  extern int my_pid;
  struct textMessage msg;
  struct Message buff;
  int bytes;

  buff.mtype = SCORE;
  buff.x = my_pid;
  bytes = send(sock, &buff, sizeof(struct Message), 0);
  if (bytes == -1) {
    perror("send");
    cleanUp(0);
  }
  bytes = recv(sock, &msg, sizeof(struct textMessage), MSG_WAITALL);
  if (msg.mtype != MESSAGE) {
    printf("Unexpected message type: %d\n", msg.mtype);
    cleanUp(0);
  }
  if (strcmp(msg.note, "Ok") != 0) {
    printf("Server says: '%s'\n", msg.note);
    cleanUp(0);
  }
}

/* draw the score on the screen */
void drawScore(WINDOW *win) {
  extern struct Table table;
  char string[80] = "";
  char temp[40];
  int i;

  for (i=0; i<MAX_PLAYERS; i++) {
    if (table.player[i].allocated) {
      sprintf(temp,"  %s: %d ", table.player[i].name, table.player[i].score);
      strcat(string,temp);
    }
  }
  mvwaddstr(win, 0, 0, "                                                                               ");
  mvwaddstr(win, 0, 40 - strlen(string)/2, string);
  sprintf(string, "%s's turn", 
	  table.player[table.turn].name);
  mvwaddstr(win, 1, 0, "                                                                               ");
  mvwaddstr(win, 1, 40 - strlen(string)/2, string);
  sprintf(string, "<j> left, <k> right, <space> stop");
  mvwaddstr(win, MAX_Y + 2, 40 - strlen(string)/2, string);
  wrefresh(win);
}

/* move the paddle based on x,y velocity and walls */
void movePaddle(int *paddlexold, int *paddlex, int movex) {

  /* move the paddle */
  *paddlexold = *paddlex;
  *paddlex = *paddlex + movex;

  /* hit side? */
  if (*paddlex - PADDLE_SIZE < MIN_X)
    *paddlex = MIN_X + PADDLE_SIZE;
  if (*paddlex + PADDLE_SIZE > MAX_X)
    *paddlex = MAX_X - PADDLE_SIZE;
}

/* move the ball based on x,y velocity and walls */
int moveBall(int *ballxold, int *ballyold, int *ballx, int *bally, 
	     int *ballx_vel, int *bally_vel, int paddlex, int paddley) {
  extern struct Table table;
  extern int other_id;
  int didScore = 0;

  /* move the ball */
  *ballxold = *ballx;
  *ballyold = *bally;
  *ballx = *ballx + *ballx_vel;
  *bally = *bally + *bally_vel; 

  /* hit side wall? */
  if (*ballx <= MIN_X || *ballx >= MAX_X) {
    *ballx_vel = *ballx_vel * -1;
  }

  /* hit paddle? */
  if (hitPaddle(ballx, ballx_vel, bally, bally_vel, paddlex, paddley))
    sendSwitchTurn();

  /* hit top? */
  if (*bally == MIN_Y + 1)
    *bally_vel = *bally_vel * -1; 

  /* Hit bottom? Then score! */
  if (*bally == MAX_Y+1) {
    sendScore();
    didScore = 1;
  }

  return didScore;


}
/* check if hit paddle */
int hitPaddle(int *ballx, int *ballx_vel, int *bally, int *bally_vel, 
	       int paddlex, int paddley) {
  int didHit = 0;

  if (*bally + *bally_vel == paddley || *bally == paddley) {

    /* left corner? */
    if (*ballx + *ballx_vel == paddlex - PADDLE_SIZE) {
      didHit = 1;
      *ballx_vel = *ballx_vel - 1;
      if (*ballx_vel < -2)
	*ballx_vel = -2;
    }

    /* right corner? */
    if (*ballx + *ballx_vel == paddlex + PADDLE_SIZE) {
      didHit = 1;
      *ballx_vel = *ballx_vel + 1;
      if (*ballx_vel > 2)
	*ballx_vel = 2;
    }

    /* middle? */
    if (((*ballx + *ballx_vel) > paddlex - PADDLE_SIZE) &&
	((*ballx + *ballx_vel) < paddlex + PADDLE_SIZE)) {
      didHit = 1;
    }

    if (didHit)
      *bally_vel = *bally_vel * -1; 

  }
  return didHit;
  
}

/* draw the screen */
void drawScreen(WINDOW *win, int ballxold, int ballyold, int ballx, int bally){
  char temp[80];
 
  mvwaddch(win, ballyold, ballxold, ' ');
  mvwaddch(win, bally, ballx, BALL_CHAR);

  drawBox(win);
  wrefresh(win);
}

void drawPaddle(WINDOW *win, int paddlexold, int paddlex, int paddley) {
  int i;

  for (i = paddlexold - PADDLE_SIZE; i <= paddlexold + PADDLE_SIZE; i++)
    mvwaddch(win, paddley, i, ' ');
  for (i = paddlex - PADDLE_SIZE; i <= paddlex + PADDLE_SIZE; i++)
    mvwaddch(win, paddley, i, PADDLE_CHAR);
}

/* draw the starting box */
void drawBox(WINDOW *win) {
  int i,j;

  /* vertical bars */
  for (i=MIN_Y; i<=MAX_Y; i++) {
    mvwaddch(win, i, MIN_X-1, '|');
    mvwaddch(win, i, MAX_X+1, '|');
  }    

  /* top bar */
  for (i=MIN_X; i<=MAX_X; i++)
    mvwaddch(win, MIN_Y, i, '-');
  mvwaddch(win, MIN_Y, MIN_X-1, 47);
  mvwaddch(win, MIN_Y, MAX_X+1, 92);
  wrefresh(win);
}

/* get keyboard move */
void getMove(WINDOW *win, int *movex) {
  int c;

  if ((c=wgetch(win)) != ERR) { /* keypressed? */
    if (c == 'j')
      *movex = -1;
    if (c == 'k')
      *movex = +1;
    if (c == ' ')
      *movex = 0;
  }
}

/* set the alarm to go off in the specified number of msecs */
void setAlarm(long milleseconds) {
  struct itimerval length;

  length.it_interval.tv_sec = 0;
  length.it_interval.tv_usec = 0;
  length.it_value.tv_sec = milleseconds/1000;
  length.it_value.tv_usec = (milleseconds % 1000) * 1000;

  if (setitimer(ITIMER_REAL,&length,NULL) < 0)
    perror("setitimer");
}

/* reset signal handler (not needed on every system) */
void onAlarm(int sig) {
  signal(SIGALRM, onAlarm);
}

/* close curses and exit */
void cleanUp(int sig) {
  extern int sock;
  struct Message buff;

  /* indicate we are quitting */
  if (sock != -1) {
    /* send exit message */
    buff.mtype = DOEXIT;
    send(sock, &buff, sizeof(struct Message), 0);
    close(sock);
  }

  /* shut curses down */
  endwin();

  /* clean up */
  printf("\nCleaned up.\n");
  exit(0);
}

/* print a usage message and quit */
void Usage(void) {
  printf("Electronic Handball, version %.2f\n", VERSION);
  printf("usage: handball {-s server} {-p #} [-h]\n");
  fprintf(stderr, "\t-p #\tport to connect to server on\n");
  fprintf(stderr, "\t-s server\tthe host running the handball server\n");
  fprintf(stderr, "\t-h\tthis usage message\n");
  exit(1);
}
